# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.flax_attention."""

import unittest

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import flax_attention
from aqt.jax_legacy.jax import flax_layers
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import test_utils
from aqt.jax_legacy.jax.flax_attention import DotProductAttnHParams
from aqt.jax_legacy.jax.flax_attention import ExpHParams
from aqt.jax_legacy.jax.flax_attention import ReciprocalHParams
from aqt.jax_legacy.jax.flax_attention import SoftmaxHParams
from aqt.jax_legacy.jax.quantization import QuantOps
from aqt.jax_legacy.jax.quantization import QuantType
from flax import jax_utils
import jax
from jax import random
from jax.nn import initializers
import jax.numpy as jnp
import numpy as onp
import tensorflow as tf

test_utils.configure_jax()


# Expressions to approximate Softmax
# Softmax with params sum_high_bound=1, low_bound=-200, clip_and_substract=True
def mod_softmax_clipped(sum_high_bound, low_bound, x, xs):
  exp_max_sub = lambda x: onp.exp(max(x, low_bound)) - onp.exp(low_bound)
  return exp_max_sub(x) / min(sum_high_bound, sum(exp_max_sub(x) for x in xs))


# Linearized exponential
def softmax_exp_lin(a, x, xs):
  exp_lin = lambda x: max(0, a * x + 1)
  return exp_lin(x) / sum(exp_lin(x) for x in xs)


# Linearized reciprocal
def softmax_recip_lin(a, x, xs):
  recip_lin = lambda x: a + 1 - a * x
  return onp.exp(x) * recip_lin(sum(onp.exp(x) for x in xs))


class SoftmaxTest(parameterized.TestCase, tf.test.TestCase):

  # Test mock input and output of modified versions of softmax.
  rand_2x2_tensor = onp.random.rand(2, 2)
  # Substract maximum from each column
  norm_dims = (0,)
  tensor_sub_max = rand_2x2_tensor - onp.max(rand_2x2_tensor, norm_dims[0])
  [[a11, a12], [a21, a22]] = tensor_sub_max

  @parameterized.named_parameters(
      dict(
          testcase_name='original_softmax',
          input_tensor=rand_2x2_tensor,
          norm_dims=norm_dims,
          softmax_hparams=SoftmaxHParams(None, None, None),
          expected_output=onp.array(
              [[
                  onp.exp(a11) / (onp.exp(a11) + onp.exp(a21)),
                  onp.exp(a12) / (onp.exp(a12) + onp.exp(a22))
              ],
               [
                   onp.exp(a21) / (onp.exp(a11) + onp.exp(a21)),
                   onp.exp(a22) / (onp.exp(a12) + onp.exp(a22))
               ]])),
      dict(
          testcase_name='exponential_clip=-200_clip_and_substract=True_' +
          'sum_high_bound=1.0',
          input_tensor=rand_2x2_tensor,
          norm_dims=norm_dims,
          softmax_hparams=SoftmaxHParams(
              exp_hparams=ExpHParams(
                  sum_high_bound=1.0,
                  low_bound=-200.0,
                  clip_and_subtract=True,
                  linear_gradient=None),
              reciprocal_hparams=None,
              quant_hparams=None),
          expected_output=onp.array(
              [[
                  mod_softmax_clipped(1.0, -200.0, a11, [a11, a21]),
                  mod_softmax_clipped(1.0, -200.0, a12, [a12, a22])
              ],
               [
                   mod_softmax_clipped(1.0, -200.0, a21, [a11, a21]),
                   mod_softmax_clipped(1.0, -200.0, a22, [a12, a22])
               ]])),
      dict(
          testcase_name='linear_gradient=1.0',
          input_tensor=rand_2x2_tensor,
          norm_dims=norm_dims,
          softmax_hparams=SoftmaxHParams(
              exp_hparams=ExpHParams(
                  sum_high_bound=0.0,
                  low_bound=0.0,
                  clip_and_subtract=False,
                  linear_gradient=1.0),
              reciprocal_hparams=None,
              quant_hparams=None),
          expected_output=onp.array([[
              softmax_exp_lin(1.0, a11, [a11, a21]),
              softmax_exp_lin(1.0, a12, [a12, a22])
          ],
                                     [
                                         softmax_exp_lin(1.0, a21, [a11, a21]),
                                         softmax_exp_lin(1.0, a22, [a12, a22])
                                     ]])),
      dict(
          testcase_name='linear_gradient=1.0.',
          input_tensor=rand_2x2_tensor,
          norm_dims=norm_dims,
          softmax_hparams=SoftmaxHParams(
              exp_hparams=ExpHParams(
                  sum_high_bound=0.0,
                  low_bound=0.0,
                  clip_and_subtract=False,
                  linear_gradient=0.0),
              reciprocal_hparams=ReciprocalHParams(
                  linear_gradient=1.0, low_bound=0.0),
              quant_hparams=None),
          expected_output=onp.array(
              [[
                  softmax_recip_lin(1.0, a11, [a11, a21]),
                  softmax_recip_lin(1.0, a12, [a12, a22])
              ],
               [
                   softmax_recip_lin(1.0, a21, [a11, a21]),
                   softmax_recip_lin(1.0, a22, [a12, a22])
               ]]),
      ),
  )
  def test_custom_softmax_vs_mock(self, input_tensor, norm_dims,
                                  softmax_hparams, expected_output):
    dtype = jnp.float32
    output = flax_attention.softmax(
        input_tensor, norm_dims, dtype, softmax_hparams,
        quant_config.DynamicContext(update_bounds=False, quantize_acts=False))
    self.assertAllClose(expected_output, output, atol=1e-6)

  # # Test modified softmax vs original softmax.
  random_input = onp.random.rand(16, 16, 2, 2)

  @parameterized.named_parameters(
      dict(
          testcase_name='modified_softmax_with_exponential_low_bound_at_-10',
          input_tensor=random_input,
          softmax_hparams=SoftmaxHParams(
              exp_hparams=ExpHParams(
                  sum_high_bound=None,
                  low_bound=-10.0,
                  clip_and_subtract=False,
                  linear_gradient=None),
              reciprocal_hparams=None,
              quant_hparams=None,
          )),
      dict(
          testcase_name='modified_softmax_with_exponential_low_bound_=-200_' +
          'clip_substract=True',
          input_tensor=random_input,
          softmax_hparams=SoftmaxHParams(
              exp_hparams=ExpHParams(
                  sum_high_bound=None,
                  low_bound=-200.0,
                  clip_and_subtract=True,
                  linear_gradient=None),
              reciprocal_hparams=None,
              quant_hparams=None),
      ),
      # Test that 'downcasting' intermediate activations to a floating-point
      # format similar to IEEE fp32 produces almost the same results as
      # unquantized softmax.
      dict(
          testcase_name='fp_downcast_fp32',
          input_tensor=random_input,
          softmax_hparams=SoftmaxHParams(
              exp_hparams=None,
              reciprocal_hparams=None,
              quant_hparams=flax_attention.SoftmaxQuantHParams(
                  reduction_prec=None,
                  prec=QuantOps.FloatQuant.FloatPrec(
                      exp_min=-2**7, exp_max=2**7, sig_bits=23))),
      ))
  def test_softmax_vs_original(self, input_tensor, softmax_hparams):
    dtype = jnp.float32
    norm_dims = (0,)
    input_tensor = jnp.array(input_tensor)
    output = flax_attention.softmax(
        input_tensor, norm_dims, dtype, softmax_hparams,
        quant_config.DynamicContext(update_bounds=False, quantize_acts=True))
    expected_output = flax_attention.softmax(
        input_tensor, norm_dims, dtype, SoftmaxHParams(None, None, None),
        quant_config.DynamicContext(update_bounds=False, quantize_acts=True))
    self.assertAllClose(expected_output, output, atol=1e-8)


class AttentionTest(parameterized.TestCase):

  @classmethod
  def construct_hparams(cls, weight_prec):
    dense = flax_layers.DenseAqt.HParams(
        weight_prec=weight_prec,
        quant_act=None,
        quant_type=QuantType.FAKE_QUANT,
        weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
        weight_half_shift=False)
    return flax_attention.MultiHeadDotProductAttentionAqt.HParams(
        dense_kqv=dense,
        dense_out=dense,
        attn_acts=DotProductAttnHParams(
            attn_act_q=None,
            attn_act_k=None,
            attn_act_probs=None,
            attn_act_v=None,
            quant_type=QuantType.FAKE_QUANT,
            softmax=SoftmaxHParams(None, None, None)))

  @parameterized.named_parameters(
      dict(testcase_name='float', weight_prec=None),
      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
  )
  def test_multihead_self_attention(self, weight_prec):
    rng = random.PRNGKey(0)
    x = jnp.ones((4, 3, 5))
    hparams = self.construct_hparams(weight_prec)
    sa_module = flax_attention.SelfAttentionAqt(
        hparams=hparams,
        num_heads=8,
        attention_axis=(1,),
        qkv_features=16,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, collect_acts_stats=False),
        train=False,
        paxis_name=None,
        kernel_init=initializers.ones,
        bias_init=initializers.zeros,
        dtype=jnp.float32,
        causal_mask=False,
        dropout_rate=0.0,
        deterministic=False,
        decode=False)
    y, _ = sa_module.init_with_output(rng, x, padding_mask=None)
    self.assertEqual(y.shape, x.shape)

  @parameterized.named_parameters(
      dict(testcase_name='float', weight_prec=None),
      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
  )
  def test_multihead_encoder_decoder_attention(self, weight_prec):
    rng = random.PRNGKey(0)
    q = jnp.ones((4, 3, 5))
    kv = jnp.ones((4, 3, 5))
    sa_module = flax_attention.MultiHeadDotProductAttentionAqt(
        num_heads=8,
        hparams=self.construct_hparams(weight_prec),
        attention_axis=(1,),
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, collect_acts_stats=False),
        train=False,
        paxis_name=None,
        qkv_features=16,
        kernel_init=initializers.ones,
        bias_init=initializers.zeros,
        dtype=jnp.float32,
        causal_mask=False,
        dropout_rate=0.0,
        deterministic=False,
        decode=False)
    y, _ = sa_module.init_with_output(
        rng, q, kv, padding_mask=None, key_padding_mask=None)
    self.assertEqual(y.shape, q.shape)

  @parameterized.named_parameters(
      dict(testcase_name='float', weight_prec=None),
      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
  )
  def test_multihead_self_attention_w_dropout(self, weight_prec):
    rng = random.PRNGKey(0)
    x = jnp.ones((4, 3, 5))
    sa_module = flax_attention.SelfAttentionAqt(
        num_heads=8,
        hparams=self.construct_hparams(weight_prec),
        attention_axis=(1,),
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, collect_acts_stats=False),
        train=False,
        paxis_name=None,
        qkv_features=16,
        kernel_init=initializers.ones,
        bias_init=initializers.zeros,
        dropout_rate=0.1,
        dtype=jnp.float32,
        causal_mask=False,
        deterministic=False,
        decode=False)
    rng_dropout, rng_params = random.split(rng)
    y, _ = sa_module.init_with_output(
        {
            'dropout': rng_dropout,
            'params': rng_params
        }, x, padding_mask=None)
    self.assertEqual(y.shape, x.shape)

  @parameterized.named_parameters(
      dict(
          testcase_name='float_spatial_shape_5_attn_dim_1',
          weight_prec=None,
          spatial_shape=(5,),
          attn_dims=(1,)),
      dict(
          testcase_name='quant_8bit_spatial_shape_5_attn_dim_1',
          weight_prec=8,
          spatial_shape=(5,),
          attn_dims=(1,)),
      dict(
          testcase_name='quant_4bit_spatial_shape_5_attn_dim_1',
          weight_prec=4,
          spatial_shape=(5,),
          attn_dims=(1,)),
  )
  def test_decoding(self, weight_prec, spatial_shape, attn_dims):
    bs = 2
    num_heads = 3
    num_features = 4
    rng = random.PRNGKey(0)
    key1, key2 = random.split(rng)
    inputs = random.normal(key1,
                           (bs,) + spatial_shape + (num_heads * num_features,))
    module = flax_attention.SelfAttentionAqt(
        num_heads=num_heads,
        hparams=self.construct_hparams(weight_prec),
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, collect_acts_stats=False),
        train=False,
        paxis_name=None,
        qkv_features=num_heads * num_features,
        attention_axis=attn_dims,
        decode=False,
        causal_mask=True,
        dtype=jnp.float32,
        dropout_rate=0.0,
        deterministic=False)

    initial_vars = module.init(key2, inputs, padding_mask=None)
    y_ref = module.apply(initial_vars, inputs, padding_mask=None)
    module.decode = True
    initial_vars_decode = module.init(key2, inputs, padding_mask=None)
    cache0 = initial_vars_decode['cache']

    def body_fn(cache, x):
      y, new_vars = module.apply({
          **initial_vars, 'cache': cache
      },
                                 x,
                                 mutable='cache',
                                 padding_mask=None)
      return new_vars['cache'], y

    # scan_in_dim supports scanning multiple dims
    _, y = jax_utils.scan_in_dim(
        body_fn, cache0, inputs, axis=attn_dims, keepdims=True)

    onp.testing.assert_allclose(y_ref, y, atol=1e-5)

  @parameterized.named_parameters(
      dict(testcase_name='float', weight_prec=None),
      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
  )
  def test_autoregresive_receptive_field_1d(self, weight_prec):
    """Tests the autoregresive self-attention receptive field."""
    rng = random.PRNGKey(0)
    rng1, rng2 = random.split(rng, num=2)

    def model_loss(inputs, pos):
      out = module.apply(initial_vars, inputs, padding_mask=None)
      assert out.shape == input_shape
      assert len(out.shape) == 3
      return out[0, pos, :].sum()

    grad_fn = jax.jit(jax.grad(model_loss))

    def get_receptive_field_1d(pos):
      g = grad_fn(inputs, pos)[0, :, :]
      return jnp.any((jnp.abs(g) > 1e-5).astype(jnp.uint32), axis=-1)

    length = 10
    dim = 1
    num_heads = 1
    input_shape = (1, length, dim)
    inputs = random.normal(rng2, input_shape)

    module = flax_attention.SelfAttentionAqt(
        num_heads=num_heads,
        hparams=self.construct_hparams(weight_prec),
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, collect_acts_stats=False),
        train=False,
        paxis_name=None,
        causal_mask=True,
        kernel_init=initializers.ones,
        dtype=jnp.float32,
        qkv_features=None,
        attention_axis=None,
        dropout_rate=0.0,
        deterministic=False,
        decode=False)
    initial_vars = module.init(
        rng1, jnp.ones((1,) + (length, dim), jnp.float32), padding_mask=None)
    # model = nn.Model(module, initial_params)

    for i in range(length):
      deps = get_receptive_field_1d(i)
      assert (deps[:i] == 1).all(), ('Receptive Field Error: Some of the '
                                     'previous positions are not reachable '
                                     'in autoregressive self-attention.')
      if i != length - 1:
        k = i + 1
        assert (deps[k:] == 0).all(), ('Receptive Field Error: Some of the '
                                       'future positions are reachable in '
                                       'autoregressive self-attention.')

  def test_padding_mask(self):
    """Test that the activation stats respect masking."""
    # This test's strategy is to change the value of a channels of a padding
    # token and make sure the stats don't change. Because the attention
    # calculation is fairly involved, this is more robust and less tedious than
    # trying to directly test numeric expected values.

    # Construct HParams with dynamic bounds.
    # Exact values don't matter, just need bounds to be dynamic so stats are
    # collected.
    bounds = get_bounds.GetBounds.Hyper(
        initial_bound=0.0,
        stddev_coeff=0.4,
        absdev_coeff=0.6,
        mix_coeff=0.4,
        reset_stats=False,
        granularity=quant_config.QuantGranularity.PER_CHANNEL)
    quant_act = flax_layers.QuantOps.ActHParams(
        input_distribution=flax_layers.QuantOps.ActHParams.InputDistribution
        .SYMMETRIC,
        prec=8,
        bounds=bounds,
        half_shift=False)
    attn_quant_act = flax_layers.QuantOps.ActHParams(
        input_distribution=flax_layers.QuantOps.ActHParams.InputDistribution
        .POSITIVE,
        prec=8,
        bounds=1.0,
        half_shift=False)
    dense_hparams = flax_layers.DenseAqt.HParams(
        quant_type=flax_layers.QuantType.FAKE_QUANT,
        weight_prec=8,
        quant_act=quant_act,
        weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
        weight_half_shift=False)
    dotproduct_attn_hparams = flax_attention.DotProductAttnHParams(
        attn_act_q=quant_act,
        attn_act_k=quant_act,
        attn_act_v=quant_act,
        attn_act_probs=attn_quant_act,
        quant_type=QuantType.FAKE_QUANT,
        softmax=SoftmaxHParams(None, None, None))
    attn_hparams = flax_attention.MultiHeadDotProductAttentionAqt.HParams(
        dense_kqv=dense_hparams,
        dense_out=dense_hparams,
        attn_acts=dotproduct_attn_hparams)

    module = flax_attention.SelfAttentionAqt(
        hparams=attn_hparams,
        num_heads=2,
        paxis_name=None,
        train=True,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=True, collect_acts_stats=False),
        dtype=jnp.float32,
        qkv_features=None,
        attention_axis=None,
        causal_mask=False,
        dropout_rate=0.0,
        deterministic=False,
        decode=False)
    # Simulate an input of a batch size of 1 with two tokens, each with four
    # features
    x = onp.arange(8).astype(onp.float32).reshape((1, 2, 4))
    initial_state = module.init(random.PRNGKey(0), x, padding_mask=None)

    padding_mask = onp.full((1, 2, 1), True)
    padding_mask[0, 1, 0] = False  # Mask out the second token
    _, state1 = module.apply(
        initial_state, x, padding_mask=padding_mask, mutable=True)
    # Now we adjust the input for the masked token and recompute the mean. It
    # should be the same as before.
    x[0, 1, 0] = 100
    _, state2 = module.apply(
        initial_state, x, padding_mask=padding_mask, mutable=True)
    test_utils.assert_stats_are_equal(state1, state2)
    # Now we adjust the input for an unmasked token and verify that the stats
    # have changed.
    x[0, 0, 0] = 200
    _, state3 = module.apply(
        initial_state, x, padding_mask=padding_mask, mutable=True)
    test_utils.assert_stats_are_unequal(state1, state3)


class AttnActsMatmulQuantTest(parameterized.TestCase):

  def construct_hparams(self, attn_act_q, attn_act_k, attn_act_probs,
                        attn_act_v):
    dense = flax_layers.DenseAqt.HParams(
        weight_prec=None,
        quant_act=None,
        quant_type=QuantType.FAKE_QUANT,
        weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
        weight_half_shift=False)
    return flax_attention.MultiHeadDotProductAttentionAqt.HParams(
        dense_kqv=dense,
        dense_out=dense,
        attn_acts=flax_attention.DotProductAttnHParams(
            attn_act_q=attn_act_q,
            attn_act_k=attn_act_k,
            attn_act_probs=attn_act_probs,
            attn_act_v=attn_act_v,
            quant_type=QuantType.FAKE_QUANT,
            softmax=SoftmaxHParams(None, None, None)))

  @parameterized.named_parameters(
      dict(
          testcase_name='float',
          attn_act_q=None,
          attn_act_k=None,
          attn_act_probs=None,
          attn_act_v=None,
          update_bounds=False,
          paxis_name=None,
          train=False),
      dict(
          testcase_name='quant_q',
          attn_act_q=QuantOps.ActHParams(
              input_distribution=QuantOps.ActHParams.InputDistribution
              .SYMMETRIC,
              prec=8,
              bounds=1,
              half_shift=False),
          attn_act_k=None,
          attn_act_probs=None,
          attn_act_v=None,
          update_bounds=False,
          paxis_name='batch',
          train=True),
      dict(
          testcase_name='quant_qk',
          attn_act_q=None,
          attn_act_k=None,
          attn_act_probs=QuantOps.ActHParams(
              input_distribution=QuantOps.ActHParams.InputDistribution
              .SYMMETRIC,
              prec=8,
              bounds=1.0,
              half_shift=False),
          attn_act_v=None,
          update_bounds=False,
          paxis_name='batch',
          train=True),
      dict(
          testcase_name='quant_k',
          attn_act_q=None,
          attn_act_k=QuantOps.ActHParams(
              input_distribution=QuantOps.ActHParams.InputDistribution
              .SYMMETRIC,
              prec=4,
              bounds=2,
              half_shift=False),
          attn_act_probs=None,
          attn_act_v=None,
          update_bounds=False,
          paxis_name=None,
          train=True),
      dict(
          testcase_name='quant_v',
          attn_act_q=None,
          attn_act_k=None,
          attn_act_probs=None,
          attn_act_v=QuantOps.ActHParams(
              input_distribution=QuantOps.ActHParams.InputDistribution
              .SYMMETRIC,
              prec=2,
              bounds=3,
              half_shift=False),
          update_bounds=True,
          paxis_name='batch',
          train=False),
      dict(
          testcase_name='quant_all_aa',
          attn_act_q=QuantOps.ActHParams(
              input_distribution=QuantOps.ActHParams.InputDistribution
              .SYMMETRIC,
              prec=8,
              bounds=1,
              half_shift=False),
          attn_act_k=QuantOps.ActHParams(
              input_distribution=QuantOps.ActHParams.InputDistribution
              .SYMMETRIC,
              prec=4,
              bounds=2,
              half_shift=False),
          attn_act_probs=QuantOps.ActHParams(
              input_distribution=QuantOps.ActHParams.InputDistribution
              .SYMMETRIC,
              prec=8,
              bounds=1.0,
              half_shift=False),
          attn_act_v=QuantOps.ActHParams(
              input_distribution=QuantOps.ActHParams.InputDistribution
              .SYMMETRIC,
              prec=2,
              bounds=3,
              half_shift=False),
          update_bounds=True,
          paxis_name=None,
          train=True),
  )
  @unittest.mock.patch.object(QuantOps, 'create_inputs_fake_quant')
  def test_self_attention_act_quant_should_call_quant_ops(
      self, mock_inputs_fake_quant, attn_act_q, attn_act_k, attn_act_probs,
      attn_act_v, update_bounds, paxis_name, train):

    mock_inputs_fake_quant.side_effect = (
        lambda inputs, hparams, bounds_params: inputs)

    rng = random.PRNGKey(0)
    x = jnp.ones((4, 3, 7))
    hparams = self.construct_hparams(attn_act_q, attn_act_k, attn_act_probs,
                                     attn_act_v)
    sa_module = flax_attention.SelfAttentionAqt(
        hparams=hparams,
        num_heads=4,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=update_bounds, collect_acts_stats=False),
        train=train,
        paxis_name=paxis_name,
        attention_axis=None,
        qkv_features=8,
        kernel_init=initializers.ones,
        bias_init=initializers.zeros,
        causal_mask=False,
        dtype=jnp.float32,
        dropout_rate=0.0,
        deterministic=False,
        decode=False)
    sa_module.init(rng, x, padding_mask=None)
    calls = []
    for hparam in [attn_act_q, attn_act_k, attn_act_probs, attn_act_v]:
      if hparam is not None:
        calls.append(
            unittest.mock.call(
                unittest.mock.ANY,
                hparams=hparam,
                bounds_params=get_bounds.GetBounds.Params(
                    update_stats=train,
                    update_bounds=update_bounds,
                    paxis_name=paxis_name,
                    mask=unittest.mock.ANY,
                    module_name=unittest.mock.ANY)))
    mock_inputs_fake_quant.assert_has_calls(calls, any_order=True)

    self.assertLen(calls, mock_inputs_fake_quant.call_count)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT tensor."""
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_tensor
import jax.numpy as jnp


# TODO(yichizh): refactor and clean the test cases
class AqtTensorTest(parameterized.TestCase):

  def test_dynamic_slice(self):
    x = jnp.array([
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [3, 5, 1, 10],
    ])
    scale = jnp.max(jnp.abs(x), axis=1, keepdims=True)
    print(x)
    print(scale)
    print(x.shape, scale.shape)

    q = aqt_tensor.QTensor(
        qvalue=x, scale=[scale], scale_t=None, dequant_dtype=scale.dtype
    )
    y = aqt_tensor.dynamic_slice(q, start_indices=(1, 0), slice_sizes=[2, 1])
    print("======")
    print(y.qvalue)
    print(y.scale[0])

  def test_getitem(self):
    x = jnp.array([
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [3, 5, 1, 10],
    ])
    scale = jnp.max(jnp.abs(x), axis=0, keepdims=True)
    print(x)
    print(scale)
    print(x.shape, scale.shape)

    q = aqt_tensor.QTensor(
        qvalue=x, scale=[scale], scale_t=None, dequant_dtype=scale.dtype
    )
    y = q.__getitem__(2)
    print("======")
    print(y.qvalue)
    print(y.scale[0])

  def test_dynamic_update(self):
    x = jnp.array([
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [3, 5, 1, 10],
    ])
    scale = jnp.max(jnp.abs(x), axis=0, keepdims=True)
    print(x)
    print(scale)
    print(x.shape, scale.shape)
    q = aqt_tensor.QTensor(
        qvalue=x, scale=[scale], scale_t=None, dequant_dtype=scale.dtype
    )

    update_qvalue = jnp.zeros((3, 1), dtype=x.dtype)
    update_scale = jnp.max(jnp.abs(update_qvalue), axis=0, keepdims=True)
    update = aqt_tensor.QTensor(
        qvalue=update_qvalue,
        scale=[update_scale],
        scale_t=None,
        dequant_dtype=update_scale.dtype,
    )
    y = aqt_tensor.dynamic_update_slice(q, update, (0, 1))
    print("======")
    print(y.qvalue)
    print(y.scale[0])


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for imagenet.train."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax.imagenet import hparams_config
from aqt.jax_legacy.jax.imagenet import models
from aqt.jax_legacy.jax.imagenet import train_utils
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4_a4_fixed
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_bfloat16
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_w4_a4_auto
from aqt.jax_legacy.utils import hparams_utils
import jax
from jax import random
import jax.numpy as jnp
import numpy as onp


class TrainTest(parameterized.TestCase):
  @parameterized.named_parameters(
      dict(
          testcase_name='quantization_none',
          base_config_filename=resnet50_bfloat16),
      dict(
          testcase_name='quantization_weights_only',
          base_config_filename=resnet50_w4),
      dict(
          testcase_name='quantization_weights_and_fixed_acts',
          base_config_filename=resnet50_w4_a4_fixed),
      dict(
          testcase_name='quantization_weights_and_auto_acts',
          base_config_filename=resnet50_w4_a4_auto),
  )  # pylint: disable=line-too-long

  def test_create_model(self, base_config_filename):
    hparams = hparams_utils.load_hparams_from_config_dict(
        hparams_config.TrainingHParams, models.ResNet.HParams,
        base_config_filename.get_config())
    model, state = train_utils.create_model(
        random.PRNGKey(0),
        8,
        224,
        jnp.float32,
        hparams.model_hparams,
        train=True)
    x = random.normal(random.PRNGKey(1), (8, 224, 224, 3))
    y, new_state = model.apply(state, x, mutable=True)
    state = jax.tree_map(onp.shape, state)
    new_state = jax.tree_map(onp.shape, new_state)
    self.assertEqual(state, new_state)
    self.assertEqual(y.shape, (8, 1000))


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General dot_general tiling."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=g-explicit-bool-comparison
# pylint: disable=g-explicit-length-test

import copy
import dataclasses
from typing import Literal
import jax
import jax.numpy as jnp
import numpy as np
from typing_extensions import Self  # for python version < 3.11


AxisIdx = int
AxisSize = int


@dataclasses.dataclass(frozen=False, slots=True)
class AxisTiling:
  axis: AxisIdx
  # At most one of tile_count, tile_size can be None.
  tile_count: AxisSize | None
  tile_size: AxisSize | None


@dataclasses.dataclass(frozen=False, slots=True)
class TensorTiling:
  contraction_axes: list[AxisTiling]
  remaining_axes: list[AxisTiling]
  # There is no point in tiling dot_general's batch axes


@dataclasses.dataclass(frozen=False, slots=True)
class Cfg:
  """Sequence of (lhs, rhs) configurations."""
  lhs: TensorTiling
  rhs: TensorTiling

  def complete_missing(
      self,
      lhs_shape: tuple[AxisSize, ...],
      rhs_shape: tuple[AxisSize, ...],
      dimension_numbers: jax.lax.DotDimensionNumbers,
  ) -> Self:
    """Makes lhs and rhs to cover all the axes."""
    new_cfg = copy.deepcopy(self)
    (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
    lhs_ra = get_ra(len(lhs_shape), lhs_ca, lhs_ba)
    rhs_ra = get_ra(len(rhs_shape), rhs_ca, rhs_ba)

    def f(axes_cfg, axes, shape):
      # add missing tile_count
      for axis_tiling in axes_cfg:
        tc = axis_tiling.tile_count
        ts = axis_tiling.tile_size
        axis_shape = shape[axis_tiling.axis]
        msg = 'At most one of tile_count and tile_size can be None'
        assert not (tc is None and ts is None), msg
        axis_tiling.tile_count = axis_shape // ts if tc is None else tc
        axis_tiling.tile_size = axis_shape // tc if ts is None else ts
        msg = (
            f'Axis {axis_tiling.axis} cannot be split into'
            f' {axis_tiling.tile_count} tiles.'
        )
        assert axis_tiling.tile_size * axis_tiling.tile_count == axis_shape, msg
      # add missing tiled axis
      axis_in_cfg = [ax.axis for ax in axes_cfg]
      for axis in axes:
        if axis not in axis_in_cfg:
          axes_cfg.append(
              AxisTiling(axis=axis, tile_count=1, tile_size=shape[axis])
          )

    f(new_cfg.lhs.contraction_axes, lhs_ca, lhs_shape)
    f(new_cfg.lhs.remaining_axes, lhs_ra, lhs_shape)
    f(new_cfg.rhs.contraction_axes, rhs_ca, rhs_shape)
    f(new_cfg.rhs.remaining_axes, rhs_ra, rhs_shape)
    return new_cfg


def interleave(ls, rs):
  assert len(ls) == len(rs)
  ret = []
  for l, r in zip(ls, rs):
    ret.append(l)
    ret.append(r)
  return ret


def zip_product(l, r):
  return map(lambda x, y: x * y, l, r)


def get_ra(rank, ca, ba) -> list[AxisIdx]:
  return list(a for a in range(rank) if a not in ca + ba)


def sort_ra_cfg(ra_tiling: list[AxisTiling]) -> list[AxisTiling]:
  ra_axes = [a.axis for a in ra_tiling]
  sorted_idx = np.argsort(ra_axes).tolist()
  return [ra_tiling[i] for i in sorted_idx]


def empty_list():
  return dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=False, slots=True)
class _Xhs:
  """Structure for bookkeeping of AxisIdx while tiling."""

  x: jnp.ndarray

  # Below are axes indices, similar to dimension_numbers
  # E.g.
  #   xhs,shape == [10, 20, 30, ...] ca=[0,1,2], ca_tile=[]
  #   After splitting axis 1 to 2 tiles, sized 10, we get
  #   xhs,shape == [10, 2, 10, 30, ...], ca=[0,2,3], ca_tile=[1]
  #   and other axes indices (ra, ra_tile, ba) also can be updated by +1.
  ca: list[AxisIdx]
  ra: list[AxisIdx]

  ca_tile: list[AxisIdx] = empty_list()  # to be summed after DG
  ra_tile: list[AxisIdx] = empty_list()  # to be reshaped after DG
  ra_tile_other: list[AxisIdx] = empty_list()  # to be reshaped after DG
  # There is no point in tiling dot_general's batch axes
  ba: list[AxisIdx] = dataclasses.field(default_factory=list)

  # axes waiting to be slitted
  # tensor_tiling: TensorTiling # TODO(lew): use this
  ca_to_be_tiled: list[AxisTiling] = empty_list()
  ra_to_be_tiled: list[AxisTiling] = empty_list()

  def tile_axis(self, at: AxisTiling, tile_axis_name: Literal['ca', 'ra']):
    """Tiles (splits) one axis while maintaining all AxisIdx."""
    shape = list(self.x.shape)
    msg = f'{shape[at.axis]=}, {at.tile_size=}, {at.tile_count=}'
    assert shape[at.axis] == at.tile_size * at.tile_count, msg
    shape[at.axis] = at.tile_size
    shape.insert(at.axis, at.tile_count)
    self.x = self.x.reshape(shape)

    def update(axes):
      for i in range(len(axes)):
        axes[i] += int(axes[i] >= at.axis)

    def update_tiling(axes):
      for i in range(len(axes)):
        axes[i].axis += int(axes[i].axis >= at.axis)

    update(self.ca)
    update(self.ca_tile)
    update(self.ra)
    update(self.ra_tile)
    update(self.ra_tile_other)
    update(self.ba)

    update_tiling(self.ca_to_be_tiled)
    update_tiling(self.ra_to_be_tiled)
    # Append tile axis index at the beginning. This has to be after the update
    match tile_axis_name:
      case 'ca':
        self.ca_tile.append(at.axis)
      case 'ra':
        self.ra_tile.append(at.axis)

  def broadcast_to_other(self, bcast_shape: tuple[AxisSize, ...]):
    """Adds new axes (bcast_shape) on AxisIdx=0."""
    assert self.ra_tile_other == list(), 'ra_tile_other already set'
    self.ra_tile_other = list(range(len(bcast_shape)))
    self.x = jnp.broadcast_to(self.x, bcast_shape + self.x.shape)

    def update(axes):
      for i in range(len(axes)):
        axes[i] += len(bcast_shape)

    update(self.ca)
    update(self.ca_tile)
    update(self.ra)
    update(self.ra_tile)
    update(self.ba)

  def axes_shape(self, axes: list[AxisIdx]) -> tuple[AxisSize, ...]:
    return tuple(map(lambda a: self.x.shape[a], axes))


def tiled_dot_general(
    cfg: Cfg,
    lhs,
    rhs,
    dimension_numbers,
    precision=None,
    preferred_element_type=None,
    dot_general=jax.lax.dot_general,
):
  """local dot_general."""

  cfg = copy.deepcopy(cfg)
  cfg = cfg.complete_missing(lhs.shape, rhs.shape, dimension_numbers)
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers

  xlhs = _Xhs(
      x=lhs,
      ca=list(lhs_ca),
      ra=get_ra(lhs.ndim, lhs_ca, lhs_ba),
      ba=list(lhs_ba),
      ca_to_be_tiled=cfg.lhs.contraction_axes,
      ra_to_be_tiled=sort_ra_cfg(cfg.lhs.remaining_axes),
  )
  xrhs = _Xhs(
      x=rhs,
      ca=list(rhs_ca),
      ra=get_ra(rhs.ndim, rhs_ca, rhs_ba),
      ba=list(rhs_ba),
      ca_to_be_tiled=cfg.rhs.contraction_axes,
      ra_to_be_tiled=sort_ra_cfg(cfg.rhs.remaining_axes),
  )

  # First tile_axis CA. CA tile_count axes will be first in xxhs.ba_tile
  assert len(xlhs.ca_to_be_tiled) == len(xrhs.ca_to_be_tiled)
  while len(xlhs.ca_to_be_tiled) > 0:
    cfg_lhs_ca = xlhs.ca_to_be_tiled.pop(0)
    cfg_rhs_ca = xrhs.ca_to_be_tiled.pop(0)
    msg = (
        'Contraction axis tile counts should be the same, but found lhs axis'
        f' {cfg_lhs_ca.axis} has a tile count of {cfg_lhs_ca.tile_count}, and'
        f' rhs axis {cfg_rhs_ca.axis} has a tile count of'
        f' {cfg_rhs_ca.tile_count}'
    )
    assert cfg_lhs_ca.tile_count == cfg_rhs_ca.tile_count, msg
    xlhs.tile_axis(cfg_lhs_ca, 'ca')
    xrhs.tile_axis(cfg_rhs_ca, 'ca')

  while len(xlhs.ra_to_be_tiled) > 0:
    cfg_lhs_ra = xlhs.ra_to_be_tiled.pop(0)
    xlhs.tile_axis(cfg_lhs_ra, 'ra')
  while len(xrhs.ra_to_be_tiled) > 0:
    cfg_rhs_ra = xrhs.ra_to_be_tiled.pop(0)
    xrhs.tile_axis(cfg_rhs_ra, 'ra')

  xlhs.broadcast_to_other(xrhs.axes_shape(xrhs.ra_tile))
  xrhs.broadcast_to_other(xlhs.axes_shape(xlhs.ra_tile))

  tiled_ca = (xlhs.ca, xrhs.ca)
  tiled_ba = (
      xlhs.ca_tile + xlhs.ba + xlhs.ra_tile + xlhs.ra_tile_other,
      xrhs.ca_tile + xrhs.ba + xrhs.ra_tile_other + xrhs.ra_tile,
  )
  new_dimension_numbers = (tiled_ca, tiled_ba)
  out = dot_general(
      xlhs.x, xrhs.x, new_dimension_numbers, precision, preferred_element_type
  )

  # Some assertions
  assert xlhs.axes_shape(xlhs.ca_tile) == xrhs.axes_shape(xrhs.ca_tile)
  ca_tile_sh = xlhs.axes_shape(xlhs.ca_tile)

  assert xlhs.axes_shape(xlhs.ba) == xrhs.axes_shape(xrhs.ba)
  ba_sh = xlhs.axes_shape(xlhs.ba)

  assert xlhs.axes_shape(xlhs.ra_tile) == xrhs.axes_shape(xrhs.ra_tile_other)
  lhs_ra_tile_sh = xlhs.axes_shape(xlhs.ra_tile)

  assert xlhs.axes_shape(xlhs.ra_tile_other) == xrhs.axes_shape(xrhs.ra_tile)
  rhs_ra_tile_sh = xlhs.axes_shape(xlhs.ra_tile_other)

  lhs_ra_sh = xlhs.axes_shape(xlhs.ra)
  rhs_ra_sh = xrhs.axes_shape(xrhs.ra)

  assert (
      out.shape
      == ca_tile_sh
      + ba_sh
      + lhs_ra_tile_sh
      + rhs_ra_tile_sh
      + lhs_ra_sh
      + rhs_ra_sh
  )

  # Sum over ca_tile now.
  # all_ba() returs ca_tile as the first axes in ba.
  assert len(xlhs.ca_tile) == len(xrhs.ca_tile)
  out = out.sum(axis=range(len(xlhs.ca_tile)))

  assert (
      out.shape
      == ba_sh + lhs_ra_tile_sh + rhs_ra_tile_sh + lhs_ra_sh + rhs_ra_sh
  )

  # Transpose tile and tile size together
  # Axis tracking is obsolete here. Only the length is the same.
  start = 0
  end_ba = len(ba_sh)
  end_lhs_ra_tile = end_ba + len(lhs_ra_tile_sh)
  end_rhs_ra_tile = end_lhs_ra_tile + len(rhs_ra_tile_sh)
  end_lhs_ra = end_rhs_ra_tile + len(lhs_ra_sh)
  end_rhs_ra = end_lhs_ra + len(rhs_ra_sh)

  from_to = lambda end1, end2: list(range(end1, end2))

  new_ba = from_to(start, end_ba)
  new_lhs_ra_tile = from_to(end_ba, end_lhs_ra_tile)
  new_rhs_ra_tile = from_to(end_lhs_ra_tile, end_rhs_ra_tile)
  new_lhs_ra = from_to(end_rhs_ra_tile, end_lhs_ra)
  new_rhs_ra = from_to(end_lhs_ra, end_rhs_ra)

  out = out.transpose(
      new_ba
      + interleave(new_lhs_ra_tile, new_lhs_ra)
      + interleave(new_rhs_ra_tile, new_rhs_ra)
  )

  lhs_ra_sh_interleaved = tuple(interleave(lhs_ra_tile_sh, lhs_ra_sh))
  rhs_ra_sh_interleaved = tuple(interleave(rhs_ra_tile_sh, rhs_ra_sh))
  lhs_ra_sh_flattened = tuple(zip_product(lhs_ra_tile_sh, lhs_ra_sh))
  rhs_ra_sh_flattened = tuple(zip_product(rhs_ra_tile_sh, rhs_ra_sh))

  assert out.shape == ba_sh + lhs_ra_sh_interleaved + rhs_ra_sh_interleaved
  out = out.reshape(ba_sh + lhs_ra_sh_flattened + rhs_ra_sh_flattened)

  return out



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AQT operation config scheduling test.

Accurate Quantized Training (AQT) operations allow parameterizing their
configuration by an event count which may change over the course of training. As
a result, different quantization configurations may be applied depending on the
value of an event count tensor containing a scalar integer "time" value.

This test ensures that the appropriate configuration is applied at the
time that the config says it will be applied.
"""

import functools

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized

from aqt.common import aqt_config
from aqt.tensorflow import aqt_ops
from aqt.tensorflow import aqt_tensor
import numpy as np
import tensorflow.compat.v1 as tf


def make_config_schedule_cases():
  # TODO(vladf): migrate this test to the class-based api once
  # conv2d and einsum have a class-based op version.
  matmul = {
      "lhs_share_stats": [0, 1],
      "rhs_share_stats": [0, 1],
      "lhs_shape": [3, 4],
      "rhs_shape": [4, 5],
      "op": aqt_ops.aqt_matmul,
      "testcase_name": "matmul"
  }

  conv2d = {
      "lhs_share_stats": [1, 2, 3],
      "rhs_share_stats": [0, 1, 2],
      "lhs_shape": [2, 5, 5, 1],
      "rhs_shape": [2, 3, 1, 1],
      "op": functools.partial(aqt_ops.aqt_conv2d, strides=1, padding="VALID"),
      "testcase_name": "conv2d"
  }

  einsum = {
      "lhs_share_stats": [0, 1, 2],
      "rhs_share_stats": [0, 2],
      "lhs_shape": [2, 1, 2, 2],
      "rhs_shape": [1, 1, 2],
      "op": functools.partial(aqt_ops.aqt_einsum, "ijkl,jmk->mil"),
      "testcase_name": "einsum"
  }

  return [matmul, conv2d, einsum]


class ConfigScheduleTest(tf.test.TestCase, parameterized.TestCase):

  @parameterized.named_parameters(make_config_schedule_cases())
  def test_config_schedule(self, lhs_share_stats, rhs_share_stats, lhs_shape,
                           rhs_shape, op):
    """Validates that the config schedule is cased appropriately.

    Construct a schedule for quantization which differs across time
    (i.e., a value of `event_count`). At every given point in time `t`,
    construct an analogous timeless quantization config that's always
    active. Ensure that the resulting quantization is identical
    for when `event_count=t` in the time-varying config.

    Note this property only holds when not freezing scale. Since quantization
    scales are stateful and update as a function of the inputs they observe,
    this test also ensures that the observed calibration statistics are
    not modified as a result of the presence of other quantization
    configurations.

    Since this test only evaluates a behavioral equality between operations with
    different configurations (timeless vs time-varying), it's agnostic to the
    actual operation being performed. For this reason, we reuse the test
    across multiple bivariate AQT operations via test case parameterization.

    Args:
      lhs_share_stats: the axes for statistics sharing on the lhs.
      rhs_share_stats: the axes for statistics sharing on the rhs.
      lhs_shape: the shape for statistics sharing on the lhs.
      rhs_shape: the shape for statistics sharing on the rhs.
      op: a callable accepting lhs/rhs quantizers and the corresponding tensors.
    """

    # For lhs, use a boring constant configuration. Note constant bound
    # coeff should cover about 67% of lhs values as they'll be standard normal.
    lhs_stats_config = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=lhs_share_stats,
        update_count_prior=0,
        tpu_cross_replica_sum=False)
    lhs_calibration_config = aqt_config.CalibrationConfig(const_bound_coeff=1)

    # For rhs, we toggle statistics updates to include/exclude zeros and
    # generate rhs from a distribution which creates zeros.
    rhs_stats_config = aqt_config.StatsConfig(
        ema_update_count=100,
        share_stats_axes=rhs_share_stats,
        update_count_prior=0,
        tpu_cross_replica_sum=False,
        l1_dev_prior=1,
        filter_zeros=True)
    rhs_calibration_config = aqt_config.CalibrationConfig(
        const_bound_coeff=1, l1_dev_coeff=1)

    def tensor_configs(lo, hi, bits):
      """Generates lhs and rhs tensor configs with specified params."""
      int_quant_config = aqt_config.IntQuantConfig(bits, preserve_zero=True)
      common_params = {
          "quant_config": int_quant_config,
          "freeze_scale_at_begin": False,
          "begin_at_event": lo,
          "end_at_event": hi
      }

      lhs_config = aqt_config.AqtTensorConfig(
          calibration_config=lhs_calibration_config,
          **common_params)
      rhs_config = aqt_config.AqtTensorConfig(
          calibration_config=rhs_calibration_config,
          **common_params)
      return [lhs_config], [rhs_config]

    def tensor_quantizer(configs, lhs_or_rhs):
      assert lhs_or_rhs in ["lhs", "rhs"], lhs_or_rhs
      stats_config = (
          lhs_stats_config if lhs_or_rhs == "lhs" else rhs_stats_config)
      config = aqt_config.AqtScheduleConfig(stats_config, configs)
      shape = lhs_shape if lhs_or_rhs == "lhs" else rhs_shape
      return aqt_tensor.TensorQuantizer(shape, config)

    def quantizer_pairs(lhs_tensor_configs, rhs_tensor_configs):
      with tf.variable_scope(None, default_name="quantizers"):
        with tf.variable_scope("lhs"):
          lhs_quantizer = tensor_quantizer(lhs_tensor_configs, "lhs")
        with tf.variable_scope("rhs"):
          rhs_quantizer = tensor_quantizer(rhs_tensor_configs, "rhs")
      return lhs_quantizer, rhs_quantizer

    lhs_ph = tf.placeholder(tf.float32, lhs_shape)
    rhs_ph = tf.placeholder(tf.float32, rhs_shape)

    def make_op(quantizers, event_count=tf.constant(0, tf.int64)):
      lhs_q, rhs_q = quantizers
      lhs_update = lhs_q.update(lhs_ph, None, event_count)
      rhs_update = rhs_q.update(rhs_ph, None, event_count)
      with tf.control_dependencies([lhs_update, rhs_update]):
        return op(lhs_q, lhs_ph, rhs_q, rhs_ph)

    # Exercise no quantization before any events occur.
    test_event_counts = [1]
    tf_expecteds = [make_op(quantizer_pairs([], []))]

    # Exercise multiple different quantizations within the schedule.
    time_varying_lhs_config, time_varying_rhs_config = [], []
    intervals = [(3, 5), (6, 7), (8, 9)]
    bits = [16, 4, 8]
    for (lo, hi), bits in zip(intervals, bits):
      lhs, rhs = tensor_configs(lo, hi, bits)
      time_varying_lhs_config += lhs
      time_varying_rhs_config += rhs

      for event_count in range(lo, hi):
        test_event_counts.append(event_count)
        tf_expecteds.append(
            make_op(quantizer_pairs(*tensor_configs(None, None, bits))))

    # Exercise no quantization before after all events occur.
    test_event_counts.append(10)
    tf_expecteds.append(make_op(quantizer_pairs([], [])))

    event_count_ph = tf.placeholder(tf.int64)
    tf_actual = make_op(
        quantizer_pairs(time_varying_lhs_config, time_varying_rhs_config),
        event_count_ph)

    # Since ops are deterministic, any nontrivial inputs will do for teasing out
    # differences. Use a half-sparse rhs to tease out zero preservation
    # configuration differences.
    rng = np.random.default_rng(1234)

    with self.cached_session():
      tf.global_variables_initializer().run()

      for i, event_count in enumerate(test_event_counts):
        lhs_value = rng.standard_normal(size=lhs_shape, dtype=np.float32)
        rhs_value = rng.standard_normal(size=rhs_shape, dtype=np.float32)
        mask = rng.choice(2, size=rhs_shape)
        rhs_value *= mask

        feed_dict = {
            event_count_ph: event_count,
            lhs_ph: lhs_value,
            rhs_ph: rhs_value
        }
        logging.info("actual eval event_count=%s", event_count)
        actual = tf_actual.eval(feed_dict)
        # Make sure to run every expected test config to update their stats,
        # even though we should only equal the i-th.
        logging.info("expected eval event_count=%s", event_count)
        expecteds = [timeless.eval(feed_dict) for timeless in tf_expecteds]
        self.assertAllEqual(actual, expecteds[i], msg=f"at event {event_count}")

  @parameterized.named_parameters(make_config_schedule_cases())
  def test_inference_schedule(self, lhs_share_stats, rhs_share_stats, lhs_shape,
                              rhs_shape, op):
    """Like test_config_schedule, but validates inference scheduling."""

    # We intentionally use constant calibration, so that we can compare
    # quantizers with different numbers of updates but identically-configured
    # serving configs.

    def make_tensor_config(lo, const_bound):
      """Generates constant-bound tensor config."""
      # Just to make use of int8 ops more variable, choose bitwidth dynamically.
      bits = 8 if const_bound <= 7 else 10
      int_quant_config = aqt_config.IntQuantConfig(
          bits=bits, preserve_zero=True)
      calibration_config = aqt_config.CalibrationConfig(
          const_bound_coeff=const_bound)
      return aqt_config.AqtTensorConfig(
          calibration_config=calibration_config,
          quant_config=int_quant_config,
          freeze_scale_at_begin=False,
          begin_at_event=lo,
          end_at_event=lo + 1)

    def make_stats_config(share_stats):
      return aqt_config.StatsConfig(
          ema_update_count=1,
          share_stats_axes=share_stats,
          tpu_cross_replica_sum=False)

    def tensor_quantizer(configs, lhs_or_rhs, inference_config_index):
      assert lhs_or_rhs in ["lhs", "rhs"], lhs_or_rhs
      share_stats = (
          lhs_share_stats if lhs_or_rhs == "lhs" else rhs_share_stats)
      stats_config = make_stats_config(share_stats)
      use_quantized_variable = False
      config = aqt_config.AqtScheduleConfig(stats_config, configs,
                                            use_quantized_variable,
                                            inference_config_index)
      shape = lhs_shape if lhs_or_rhs == "lhs" else rhs_shape
      return aqt_tensor.TensorQuantizer(shape, config)

    def quantizer_pairs(lhs_configs, rhs_configs, inference_index):
      with tf.variable_scope(None, default_name="quantizers"):
        with tf.variable_scope("lhs"):
          lhs_quantizer = tensor_quantizer(lhs_configs, "lhs", inference_index)
        with tf.variable_scope("rhs"):
          rhs_quantizer = tensor_quantizer(rhs_configs, "rhs", inference_index)
      return lhs_quantizer, rhs_quantizer

    lhs_tensor_configs = [make_tensor_config(i, i) for i in range(1, 10)]
    rhs_tensor_configs = [make_tensor_config(i, 10 - i) for i in range(1, 10)]
    dynamic_quantizers = quantizer_pairs(lhs_tensor_configs, rhs_tensor_configs,
                                         None)
    static_quantizerss = [
        quantizer_pairs(lhs_tensor_configs, rhs_tensor_configs, i - 1)
        for i in range(1, 10)
    ]

    lhs_ph = tf.placeholder(tf.float32, lhs_shape)
    rhs_ph = tf.placeholder(tf.float32, rhs_shape)
    event_count_ph = tf.placeholder(tf.int64, [])

    def update_quantizers_op(quantizers):
      lhs_q, rhs_q = quantizers
      return tf.group([
          lhs_q.update(lhs_ph, None, event_count_ph),
          rhs_q.update(rhs_ph, None, event_count_ph)
      ])

    update_dynamic_quantizers = update_quantizers_op(dynamic_quantizers)
    update_static_quantizerss = [
        update_quantizers_op(static_quantizers)
        for static_quantizers in static_quantizerss
    ]

    def make_serving_op(quantizers):
      lhs_q, rhs_q = quantizers
      return op(lhs_q, lhs_ph, rhs_q, rhs_ph, train=False)

    dynamic_op = make_serving_op(dynamic_quantizers)
    static_ops = [
        make_serving_op(static_quantizers)
        for static_quantizers in static_quantizerss
    ]

    rng = np.random.default_rng(1234)

    with self.cached_session() as sess:
      tf.global_variables_initializer().run()
      lhs_value = rng.standard_normal(size=lhs_shape, dtype=np.float32)
      rhs_value = rng.standard_normal(size=rhs_shape, dtype=np.float32)

      # Initialize dynamic quantizer at time 0.
      feed_dict = {lhs_ph: lhs_value, rhs_ph: rhs_value, event_count_ph: 0}
      sess.run(update_dynamic_quantizers, feed_dict)

      for i in range(1, 10):
        lhs_value = rng.standard_normal(size=lhs_shape, dtype=np.float32)
        rhs_value = rng.standard_normal(size=rhs_shape, dtype=np.float32)
        feed_dict = {lhs_ph: lhs_value, rhs_ph: rhs_value, event_count_ph: i}
        sess.run(update_dynamic_quantizers, feed_dict)
        sess.run(update_static_quantizerss[i - 1], feed_dict)

        dynamic_result = sess.run(dynamic_op, feed_dict)
        static_result = sess.run(static_ops[i - 1], feed_dict)
        self.assertAllEqual(
            dynamic_result,
            static_result,
            msg=f"dynamic event {i} static event {i}")

        # Invasively update the quantizers' last update.
        for q in static_quantizerss[i - 1]:
          sess.run(q._last_update.assign(tf.zeros_like(q._last_update)))

        static_result = sess.run(static_ops[i - 1], feed_dict)
        self.assertAllEqual(
            dynamic_result,
            static_result,
            msg=f"dynamic event {i} static event 0")


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=g-importing-member
"""Utilities for managing the hyperparameter configurations of experiments."""

import dataclasses
import typing
from typing import Optional, TypeVar

from absl import flags

from aqt.jax_legacy.jax.flax import struct as flax_struct
from aqt.jax_legacy.jax.wmt_mlperf import models
from aqt.jax_legacy.utils import hparams_utils as os_hparams_utils

T = TypeVar('T')

FLAGS = flags.FLAGS

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass

# TODO(malmaud): Have these be subclasses of a new 'Train' class that implements
# the training loop, in analogy to how Transformer.HParams is a subclass of
# Transformer. That will be a non-trivial refactoring of train.py, so
# postpone for a follow-up CL.


@dataclass
class LearningRateSchedulerHParams:
  """The set of hyperparameters used to control the learning rate scheduler.

  See train.create_learning_rate_scheduler for the corresponding
  implementation.
  """

  # A string with factors separated by '*' that defines the schedule.
  # Consists of:
  # * constant: interpreted as the constant value,
  # * linear_warmup: interpreted as linear warmup until warmup_steps,
  # * rsqrt_decay: divide by square root of max(step, warmup_steps)
  # * decay_every: Every k steps decay the learning rate by decay_factor.
  # * cosine_decay: Cyclic cosine decay, uses steps_per_cycle parameter.
  # For example, "constant * linear_warmup * rsqrt_decay".
  factors: str

  # the starting constant for the lr schedule.
  base_learning_rate: float

  # how many steps to warm up for in the warmup schedule.
  warmup_steps: int

  # The amount to decay the learning rate by.
  decay_factor: Optional[float]

  # How often to decay the learning rate.
  steps_per_decay: Optional[int]

  # Steps per cycle when using cosine decay.
  steps_per_cycle: Optional[int]


# TODO(b/176825799): Create shared training hparams for different models i.e.
# wmt_mlperf and imagenet.
@dataclass
class TrainingHParams:
  """The set of hyperparameters used by the training loop."""

  metadata: os_hparams_utils.HParamsMetadata

  # Optimizer parameters
  learning_rate_schedule: LearningRateSchedulerHParams
  per_host_batch_size: int
  num_train_steps: int
  weight_decay: float  # Decay factor for AdamW style weight decay
  beta1: float
  beta2: float
  eps: float

  # RNG parameters
  random_seed: int  # Integer for PRNG random seed
  hardware_rng: bool  # Whether to use hardware rng for dropout

  # Auto clip activation quantization parameter
  activation_bound_update_freq: int  # Update frequency. If set to a value > 0,
  # will set update_bounds to True at step = 1 * freq, step = 2 * freq, etc..
  # If set to a value <= 0, update_bounds will always be False.
  # If GetBounds is used anywhere in the model, then this parameter should be
  # set to a value > 0.
  # TODO(wanglisa): Validate this before training starts (b/162350655)

  activation_bound_start_step: int

  prefer_int8_to_int32_dot: bool  # lax.dot inputs with an int8 dtype.

  # Weight outlier regularization parameters
  weight_outlier_regularization: float
  weight_outlier_regularization_regex: str

  model_hparams: models.Transformer.HParams

  @classmethod
  def from_config_dict(cls, config_dict):
    return os_hparams_utils.load_dataclass_from_config_dict(cls, config_dict)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates a sweep over multiple sized models.

Sweep:
[{number of layers x emb dim x (heads,  qkv,  mlp)} x quantization x sparsity]
Sweep for {3, 6} x {1024} x {(8, 512, 1024), (8, 1024, 2048), (16, 512, 2048),
(16, 1024, 4096)}
"""

import copy
import itertools

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs import base_config


def get_model_size_configs():
  """Sweeps over multiple num_layers, heads dims, qkv_dims and mlp_dims."""
  configs = []

  for num_layers, (num_heads, qkv_dim, mlp_dim) in itertools.product(
      [3, 6],
      [(8, 512, 1024), (8, 1024, 2048), (16, 512, 2048), (16, 1024, 4096)],
  ):
    base_config_file = base_config.get_config(
        n_layers=num_layers,
        quant_target=base_config.QuantTarget.WEIGHTS_AND_AUTO_ACTS,
    )
    config = copy.deepcopy(base_config_file)
    config.metadata.hyper_str = f"wmt_L({num_layers})_H({num_heads})_QKV({qkv_dim})_MLP({mlp_dim})_emb(1024)"
    model = config.model_hparams
    model.num_heads = num_heads
    model.qkv_dim = qkv_dim
    model.mlp_dim = mlp_dim
    configs.append(config)

  return configs



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.summary_utils."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax.stats import Stats
from aqt.jax_legacy.utils import summary_utils
import jax.numpy as jnp
import numpy as np


class SummaryUtilsTest(parameterized.TestCase):

  def assertNestedDictEqual(self, a, b):
    np.testing.assert_equal(a.keys(), b.keys())
    for key in a:
      np.testing.assert_array_equal(a[key], b[key])

  @parameterized.named_parameters(
      dict(testcase_name='no_keys', keys=[]),
      dict(testcase_name='key_not_in_dict', keys=['bounds']))
  def test_empty_get_state_dict_summary(self, keys):
    state_dict = {}
    distr_summary = summary_utils.get_state_dict_summary(state_dict, keys=keys)
    self.assertEmpty(distr_summary)

  @parameterized.named_parameters(
      dict(
          testcase_name='keys_in_dict',
          keys=['bounds', 'min_per_ch', 'max_per_ch'],
          expected_summary={
              '/decoder/attention/dense_out/bounds':
                  np.array([[1., 2.], [2., 4.], [3., 6.]]),
              '/decoder/attention/dense_out/min_per_ch':
                  np.array([-6., -5., -4.]),
              '/decoder/attention/dense_out/max_per_ch':
                  np.array([20., 21., 22.]),
          }),
      dict(
          testcase_name='key_not_in_dict',
          keys=['other_key'],
          expected_summary={}))
  def test_get_state_dict_summary(self, keys, expected_summary):
    state_dict = {
        'decoder': {
            'attention': {
                'dense_out': {
                    'bounds':
                        jnp.array([[1., 2.], [2., 4.], [3., 6.]]),
                    'min_per_ch':
                        jnp.array([-6., -5., -4.]),
                    'max_per_ch':
                        jnp.array([20., 21., 22.]),
                    'stats':
                        Stats(
                            n=1,
                            mean=jnp.ones(()),
                            mean_abs=jnp.ones(()),
                            mean_sq=jnp.ones(()),
                            mean_batch_maximum=jnp.ones(()),
                            mean_batch_minimum=jnp.ones(()))
                }
            },
            'mlp': {
                'dense_1': {
                    'stats':
                        Stats(
                            n=1,
                            mean=jnp.ones(()),
                            mean_abs=jnp.ones(()),
                            mean_sq=jnp.ones(()),
                            mean_batch_maximum=jnp.ones(()),
                            mean_batch_minimum=jnp.ones(()))
                }
            },
        }
    }

    summary = summary_utils.get_state_dict_summary(state_dict, keys=keys)
    self.assertNestedDictEqual(summary, expected_summary)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataclass implementation of AQT configuration."""

import dataclasses
import enum
from typing import Any, Dict, List, Optional, Type, Union

from aqt.common.aqt_config_utils import _BaseConfig
from aqt.common.aqt_config_utils import _validate_intervals
from aqt.common.aqt_config_utils import ConfigError



@dataclasses.dataclass
class IntQuantConfig(_BaseConfig):
  # pyformat: disable
  """Config for integer quantization in aqt_ops.

  Attributes:
    bits:
      Number of bits to quantize to (e.g 4 for int4). Must be positive.
    preserve_zero:
      Whether or not zeros should be representable.
  """
  # pyformat: enable
  bits: int
  preserve_zero: bool = True

  def validate(self):
    if self.bits < 1:
      raise ConfigError(f'expected bits={self.bits} > 0')

  def compatible_with_int8(self) -> bool:
    """Checks if the config is compatible with int8."""
    return self.bits <= 8


@enum.unique
class RoundingMode(enum.Enum):
  """Supported rounding mode for SmallFloatConfig."""
  ROUND_AWAY_FROM_ZERO = 'round_away_from_zero'
  ROUND_TO_NEAREST_EVEN = 'round_to_nearest_even'
  ROUND_STOCHASTIC = 'round_stochastic'


@dataclasses.dataclass
class SmallFloatConfig(_BaseConfig):
  """Enum for emulated lower-precision formats."""
  exponent_bits: int
  mantissa_bits: int
  min_exp: int
  max_exp: int
  support_inf: bool  # Not yet supported
  rounding_mode: RoundingMode  # Currently only using ROUND_TO_NEAREST_EVEN


@dataclasses.dataclass
class FloatConfig(_BaseConfig):
  """Config for non quantization in aqt_ops.

  Attributes:
    use_bf16: Whether or not to use bfloat16.
  """
  use_bf16: bool = True


# Warning: Adding new oneof fields may cause from_dict to select the wrong
# value. The behavior of dataclasses is to select the first value in the
# union which has all values specified For example if FloatConfig were first,
# it would always be selected since its only field use_bf16 has a default
# value.
QuantConfig = Union[IntQuantConfig, SmallFloatConfig, FloatConfig]


@dataclasses.dataclass
class StatsConfig(_BaseConfig):
  # pyformat: disable
  """Config for aqt_ops.Stats.

  Attributes:
    ema_update_count:
      Scale calibration sample size.
      The weight of the newest batch will be 1 / ema_update_count.
      E.g. 1 means that only the current batch is used in calibration.
      Weights of examples in a batch are also correctly taken into account.
    share_stats_axes:
      Bound sharing. Required for all contraction axes. Each axis in this list
      must be in [0, rank) where rank is the rank of the tensor to quantize.
      This list must also be strictly sorted.
    filter_zeros:
      Ignores zeros when computing statistics. Useful for distributions with a
      zero spike like the output of Relu.
    lp_order:
      Defines moment used in lp_dev, needs to be even and positive.

    update_count_prior:
      Emulates a situation where some statistics have already been collected.
      Default of `1.0` prevents division by zero in prod, while setting it to
      `0` allows for precise logic in unit testing.
    mean_prior:
      Emulates a situation where some statistics have already been collected.
    l1_dev_prior:
      Emulates a situation where some statistics have already been collected.
    lp_dev_prior:
      Emulates a situation where some statistics have already been collected.
    max_dev_prior:
      Encode prior knowledge of the maximum absolute value observed; note that
      the `update_count_prior` does not affect the prior weight for the max so
      in practice this may be usually left at zero.

    tpu_cross_replica_sum:
      Setting it to false might improve performance, but note that checkpoint
      might contain wrong value.

    safe_divide:
      When any components in the input tensor are zeros along the shared stats
      axes (or all weights are zeros) and filter_zeros is True and
      ema_update_count is 1 with no moving average, the sum_of_ones will have a
      zero component. In such case, use divide_no_nan when updating stats.
  """
  # pyformat: enable
  ema_update_count: int
  share_stats_axes: List[int]
  filter_zeros: bool = True
  lp_order: int = 2

  update_count_prior: float = 1.0
  mean_prior: float = 0.0
  l1_dev_prior: float = 0.0
  lp_dev_prior: float = 0.0
  max_dev_prior: float = 0.0

  tpu_cross_replica_sum: bool = True

  safe_divide: bool = False

  def validate(self, data_shape: List[Optional[int]],  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
               dynamic: bool = False):
    """Validates this StatsConfig for the provided data shape.

    Args:
      data_shape: the shape of the input tensor which will be quantized with
        self as the statistics configuration. If an entry is None, this
        indicates a dimension whose size is unknown at graph compilation time.
      dynamic: whether the quantization is dynamic.

    Raises:
      ConfigError: if any of the specified share_stats_axes are not between
        [0, rank) where rank is len(data_shape), the share_stats_axes are
        not sorted, don't contain all unknown dimensions (None in data_shape),
        or if ema_update_count is not positive.
    """
    rank = len(data_shape)
    if any(ax < 0 or ax >= rank for ax in self.share_stats_axes):  # pylint: disable=not-an-iterable
      raise ConfigError(f'share_stats_axes ({self.share_stats_axes}) must be '
                        f'between 0 and rank ({rank}), not including rank')
    if not all(ax1 < ax2 for ax1, ax2 in zip(self.share_stats_axes[:-1],
                                             self.share_stats_axes[1:])):
      raise ConfigError(
          f'share_stats_axes ({self.share_stats_axes}) must be strictly sorted')

    if not dynamic:
      unknown_axes = {i for i, dim in enumerate(data_shape) if dim is None}
      shared_axes = set(self.share_stats_axes)
      if not unknown_axes.issubset(shared_axes):
        raise ConfigError(f'expected share_stats_axes ({self.share_stats_axes})'
                          ' to contain unknown axes for given data shape '
                          f'({data_shape})')

    if self.ema_update_count < 1:
      raise ConfigError(
          f'expected ema_update_count={self.ema_update_count} >= 1')

    if self.lp_order < 1:
      raise ConfigError(f'expected lp_order={self.lp_order} >= 1')

    if self.safe_divide and not self.filter_zeros:
      raise ConfigError(f'expected safe_divide={self.safe_divide} is False, '
                        'no need to use safe_divide since sum_of_ones is '
                        'non-zero when filter_zeros is False')


@dataclasses.dataclass
class CalibrationConfig(_BaseConfig):
  """Config for aqt_ops.Stats.

  Attributes:
    const_bound_coeff: Bias used in bound calibration.
    l1_dev_coeff: Weight of the l1_dev in bound calibration.
    lp_dev_coeff: Weight of the lp_dev in bound calibration.
    max_dev_coeff: Weight of the max_dev in bound calibration.
  """
  const_bound_coeff: float = 0.0
  l1_dev_coeff: float = 0.0
  lp_dev_coeff: float = 0.0
  max_dev_coeff: float = 0.0


@dataclasses.dataclass
class AqtTensorConfig(_BaseConfig):
  """Config for aqt_ops.AqtTensor.

  Attributes:
    quant_config: Numerical format for quantization.
    calibration_config: Calibration config.
    freeze_scale_at_begin: Freezes the calibration when begin_at_event is
      reached. This breaks the feedback look between calibration scale and
      distribution observed by stats.
    begin_at_event: Start of this quantization interval. You need to make sure
      that there are enough updates before quantization is enabled with
      begin_at_event or set a proper prior or use const_bound. This is very
      important if freeze_scale_at_begin == true.
    end_at_event: End of this quantization interval.
  """
  quant_config: QuantConfig
  calibration_config: CalibrationConfig
  freeze_scale_at_begin: bool
  begin_at_event: Optional[int] = None
  end_at_event: Optional[int] = None

  def validate(self):
    """Validates this tensor config."""
    if (self.begin_at_event is not None and self.end_at_event is not None and
        self.begin_at_event > self.end_at_event):
      raise ConfigError(f'expected begin_at_event={self.begin_at_event} <= '
                        f'end_at_event={self.end_at_event}')

    if not isinstance(self.quant_config, IntQuantConfig) and not isinstance(
        self.quant_config, FloatConfig) and not isinstance(
            self.quant_config, SmallFloatConfig):
      raise ConfigError('quant_config must be one of '
                        '{int_quant_config, float_config, small_float_config}.')
    if isinstance(self.quant_config, SmallFloatConfig):
      small_float = self.quant_config
      if small_float.support_inf:
        raise NotImplementedError('support_inf is not yet supported.')
      if small_float.rounding_mode != RoundingMode.ROUND_TO_NEAREST_EVEN:
        raise NotImplementedError(
            'Using small_float.rounding_mode '
            'only currently supports ROUND_TO_NEAREST_EVEN.')

  def to_dict(self) -> Dict[str, Any]:
    # AqtTensorConfig dataclass does not have `int_quant_config` and
    # `float_config` which are present in proto. Instead, it comes with
    # `quant_config`, but it is not a valid field in the corresponding proto.
    # So, they are manually added/removed below for proper conversion.
    dataclass_dict = dataclasses.asdict(self)
    cfg_type = 'int_quant_config' if isinstance(
        self.quant_config, IntQuantConfig) else 'float_config'
    dataclass_dict[cfg_type] = dataclass_dict.pop('quant_config')
    return dataclass_dict


@dataclasses.dataclass
class AqtScheduleConfig(_BaseConfig):
  """Config for an `aqt_tensor.TensorQuantizer` configuration schedule.

  Attributes:
    stats_config: Statistics configs describing what statistics to track.
    tensor_configs: Tensor configs to apply to the input tensor. *_at_event must
      describe sorted and disjoint intervals. *_at_event must be identical in
      the filter configs.
    use_quantized_variable: If true and in training mode, saves intermediate
      quantizations to user-provided variables. During inference, quantized
      variables are read from but not written to with new input values.
    inference_config_index: If not None, then the index into `tensor_configs`
      which indicates which tensor configuration to use during inference. This
      allows for static switching logic at inference time, which improves
      throughput. If None, then the most recently trained-on config is the one
      that's used.
    allow_int_small_float: Whether Integer and SmallFloat configs can exist in
      the same schedule. This field allows later checks to be done at the Python
      level instead of introducing additional tf.where operations.
  """
  stats_config: StatsConfig
  tensor_configs: List[AqtTensorConfig]
  use_quantized_variable: bool = False
  inference_config_index: Optional[int] = None
  allow_int_small_float: Optional[bool] = None

  def quantization_mode(self) -> Type[QuantConfig]:
    """Returns which quantization to use.

    Raises: ConfigError if both IntQuantConfig and SmallFloatConfig are given.
    """
    has_quantization = False
    has_small_float = False
    for tensor_config in self.tensor_configs:
      quant_config = tensor_config.quant_config
      if isinstance(quant_config, IntQuantConfig):
        has_quantization = True
      elif isinstance(quant_config, SmallFloatConfig):
        has_small_float = True
      elif isinstance(quant_config, FloatConfig):
        continue
      else:
        raise ConfigError(
            'quant_config must be one of '
            '{int_quant_config, float_config, small_float_config}.')
    if has_quantization and has_small_float:
      raise ConfigError(
          'Found both IntQuantConfig and SmallFloatConfig. Only '
          'one of the two should be specified among all tensor_configs.')
    elif has_quantization:
      return IntQuantConfig
    elif has_small_float:
      return SmallFloatConfig
    else:
      return FloatConfig

  def validate(self, data_shape: List[Optional[int]]):  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    """Validates this AqtScheduleConfig for the provided data shape."""
    # The output value of quantization_mode is unused.
    if not self.allow_int_small_float:
      self.quantization_mode()
    _validate_intervals(self.tensor_configs)
    if any(ax is None for ax in data_shape) and self.use_quantized_variable:
      raise ConfigError(
          'use of quantized variable with unknown dimensions is disallowed')
    if self.inference_config_index is not None and not (
        0 <= self.inference_config_index < len(self.tensor_configs)):
      raise ConfigError(
          f'inference_config_index ({self.inference_config_index}) must be '
          'at least 0 and less than len(tensor_configs) '
          f'({len(self.tensor_configs)})')

  def fill_gaps_with_float_config(self):
    """Fills gaps with FloatConfig to always have one active config."""

    def create_float_config(begin, end):
      tc = AqtTensorConfig(
          quant_config=FloatConfig(),
          calibration_config=CalibrationConfig(),
          freeze_scale_at_begin=True,
          begin_at_event=begin,
          end_at_event=end)
      return tc

    previous_end = None
    filled_configs = []
    inference_config_index = None
    for i, config in enumerate(self.tensor_configs):
      current_begin = config.begin_at_event

      if previous_end != current_begin and current_begin > 0:
        filled_configs.append(create_float_config(previous_end, current_begin))

      filled_configs.append(config)
      previous_end = config.end_at_event

      if self.inference_config_index == i:
        inference_config_index = len(filled_configs) - 1

    if previous_end is not None or not filled_configs:
      filled_configs.append(create_float_config(previous_end, None))

    self.tensor_configs = filled_configs
    # Note that self.inference_config_index is either None or in [0, len)
    # where len is the length of the previous self.tensor_configs. So the
    # update below either keeps it set to None or updates the index to the
    # reindexed value for the new self.tensor_configs.
    self.inference_config_index = inference_config_index


@dataclasses.dataclass
class AqtMatmulConfig(_BaseConfig):
  """Quantization config for a matmul.

  Attributes:
    lhs: quantization schedule for left-hand side argument
    rhs: quantization schedule for right-hand side argument
    grad: quantization schedule for gradients (optional).
  """
  lhs: AqtScheduleConfig
  rhs: AqtScheduleConfig
  grad: Optional[AqtScheduleConfig] = None


@dataclasses.dataclass
class AqtEinsumConfig(_BaseConfig):
  """Quantization config for a two-argument einsum.

  Attributes:
    lhs: quantization schedule for left-hand side argument
    rhs: quantization schedule for right-hand side argument
    lhs_grad: quantization schedule for gradients used to calculate gradients of
      left-hand side argument.
    rhs_grad: quantization schedule for gradients used to calculate gradients of
      right-hand side argument.
  """
  lhs: AqtScheduleConfig
  rhs: AqtScheduleConfig
  lhs_grad: Optional[AqtScheduleConfig] = None
  rhs_grad: Optional[AqtScheduleConfig] = None




# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataclass implementation of AQT configuration."""

import dataclasses
import enum

from typing import Any, Dict, List, Type, TypeVar

import dacite


class ConfigError(Exception):
  pass


T = TypeVar('T', bound='_BaseConfig')


@dataclasses.dataclass
class _BaseConfig:
  """Adds serialization and primitive type checking to dataclasses."""

  def to_dict(self) -> Dict[str, Any]:
    """Converts a dataclass to dict while preserving dataclass-specific logic."""
    dataclass_dict = {}
    for field in dataclasses.fields(self):
      attr = getattr(self, field.name)
      if isinstance(attr, _BaseConfig):
        dataclass_dict[field.name] = attr.to_dict()
      elif isinstance(attr, list):
        dataclass_dict[field.name] = [
            a.to_dict() if isinstance(a, _BaseConfig) else a for a in attr
        ]
      # For enums pass the name to use for making the proto.
      elif isinstance(attr, enum.Enum):
        dataclass_dict[field.name] = attr.name
      else:
        dataclass_dict[field.name] = attr

    return dataclass_dict

  @classmethod
  def from_dict(cls: Type[T], data: Dict[str, Any]) -> T:
    return dacite.from_dict(cls, data)

  def validate(self):
    pass


# TODO(lew): Once the dependencies are untangled, we can uncomment this line.
# aqt_tensor_config_type = 'AqtTensorConfig'
aqt_tensor_config_type = Any


def _validate_intervals(configs: List[aqt_tensor_config_type]):
  """Ensures that the provided schedule specifies increasing time segments.

  We require configs to be in sorted order with respect to
  `(start_at_event, end_at_event)` intervals. This makes disjointness checking
  easier.

  Args:
    configs: a list of tensor quantization configs.

  Raises:
    ConfigError: if any of the described conditions are not met.
  """
  previous_end = None
  for i, config in enumerate(configs):
    current_begin = config.begin_at_event
    current_end = config.end_at_event

    neither_none = current_begin is not None and current_end is not None
    if neither_none and current_end < current_begin:
      raise ConfigError(f'config[{i}].end_at_event ({current_end}) < '
                        f'config[{i}].begin_at_event ({current_begin})')

    if i > 0:
      if current_begin is None:
        raise ConfigError('only the first config may omit begin_at_event, but '
                          f'config[{i}] does too')

      if previous_end is None:
        raise ConfigError('only the last config may omit end_at_event, but '
                          f'config[{i-1}] does too')

      if previous_end > current_begin:
        raise ConfigError(f'config[{i-1}].end_at_event ({previous_end}) > '
                          f'config[{i}].begin_at_event ({current_begin})')

    if previous_end is not None and previous_end != current_begin:
      raise ConfigError(
          f'There is a gap between config[{i-1}].end_at_event ({previous_end}) '
          f'and config[{i}].begin_at_event ({current_begin}). There must be '
          f'always one active config.')

    previous_end = current_end


def _validate_alignment(
    lhs_path: str,  #
    lhs_configs: List[aqt_tensor_config_type],
    rhs_path: str,
    rhs_configs: List[aqt_tensor_config_type]):
  """Ensures that quantization schedules for both arguments are aligned.

  For binary quantized operations, a quantization schedule for each argument
  specifies the type of quantization that should be applied at a given time.

  The configuration is assumed to specify equal-length lists of quantization
  configurations for LHS and RHS configs, for which parallel entries must
  have identical endpoints for quantization activation.

  Args:
    lhs_path: what to call the `lhs` in error strings.
    lhs_configs: the tensor quantization schedule for the left argument.
    rhs_path: what to call the `rhs` in error strings.
    rhs_configs: the tensor quantization schedule for the right argument.

  Raises:
    ConfigError: if any of the described conditions are not met.
  """
  if len(lhs_configs) != len(rhs_configs):
    raise ConfigError(f'lhs config len {len(lhs_configs)} must equal '
                      f'rhs config len {len(rhs_configs)}')

  for i, (lhs_config, rhs_config) in enumerate(zip(lhs_configs, rhs_configs)):

    def _interval(config):
      return config.begin_at_event, config.end_at_event

    if _interval(lhs_config) != _interval(rhs_config):
      raise ConfigError(
          '(begin_at_event, end_at_event) intervals do not match: '
          f'{lhs_path}[{i}] ({_interval(lhs_config)}) != '
          f'{rhs_path}[{i}] ({_interval(rhs_config)})')



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions shared by modules in aqt/jax."""


def normalize_axes(axes, ndim):
  # A tuple by convention. len(axes_tuple) then also gives the rank efficiently.
  return tuple([ax if ax >= 0 else ndim + ax for ax in axes])


def broadcast_rank(source, target):
  """Broadcasts source to match target's rank following Numpy semantics."""
  return source.reshape((1,) * (target.ndim - source.ndim) + source.shape)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""utils contains all kinds of small pieces of code used throughout AQT.

Code in this file can't depend on any other AQT file.
However, it is acceptable to grow pieces of funcionality in this file and later
promote them to dedicated files.
"""

import difflib
import functools
import pprint
from typing import Any
import flax.struct


flax_slots_dataclass = functools.partial(
    flax.struct.dataclass, frozen=False, slots=True
)


def static_field(**kwargs):
  return flax.struct.field(pytree_node=False, **kwargs)


def dynamic_field(**kwargs):
  return flax.struct.field(pytree_node=True, **kwargs)


def print_diff(str_a: str, str_b: str):
  diff_generator = difflib.context_diff(str_a.split(' '), str_b.split(' '))
  for diff in diff_generator:
    print(diff)


def test_pprint_eq(input_a: Any, input_b: Any):
  str_input_a = input_a if isinstance(input_a, str) else pprint.pformat(input_a)
  str_input_b = input_b if isinstance(input_b, str) else pprint.pformat(input_b)
  assert str_input_a == str_input_b, print_diff(str_input_a, str_input_b)
